# ASLR-Analysis

To evaluate ASLR effectiveness we need to first collect the sample and then perform the analysis on the data.
We consider the following operating systems:

* Linux (Ubuntu 22.04) - Kernel 5.17.15 - x86_64
* Linux (Ubuntu 22.04) - Kernel 6.4.9 - x86_64
* MacOS (Ventura 13.4.1) - ARM M1
* MacOS (Ventura 13.4.1) - Rosetta framework
* Windows (Version 11) - x86_64

## Sampling

Sampling on Linux can be done by executing the python3 script `src/sampling/main.py`. Hence:

```
cd src/sampling/
python3 main.py
```

This script allows to collect the provided number of samples in one boot and all the data will be saved in the directory `src/sampling/data_<os>/archives`.
However, in MacOS and Windows we have sections that are randomized at boot time.
In these cases, we need a script that calls `main.py` at boot and reboots the machine after collecting a given amount of samples.
The reboot scripts are already available in the directory `src/sampling/REBOOTER`.
Please refer to the README in these directories for the reboot setup.

## Paper dataset

The dataset of the paper can be found at: https://zenodo.org/records/12784924

The structure of the dataset is the following:

```
- samples
  - data_linux
    - 5.17.15
	    - archives
    - 6.4.9
	    - archives
  - data_macos
    - native
	    - singleboot
		    - archives
	    - reboot
		    - archives
    - rosetta
	    - singleboot
		    - archives
	    - reboot
		    - archives
  - data_windows
    - singleboot
	    - archives
    - reboot
	    - archives
```

The directories `reboot` should contain the samples obtained with the rebooter scripts, while the directories `singleboot` should contain the samples obtained with the `main.py` script.

## Analysis

The scripts for the analysis are in the directory `src/analysis`. It is highly recommended to install all the dependencies in `requirements.txt` within a virtual environment.
To decide which OS samples to analyze, edit the file `src/analysis/config.py`.
The first two scripts, namely `0_archives_to_csv.py` and `1_Preprocessing.ipynb`, must be executed in order. Afterward, it is possible to run the other script to obtain the results.

Note: these scripts consider the aforementioned filesystem structure. Changes in the filesystem structure will result in errors.

## Sudoedit (CVE-2023-22809) PoC

PoC available at: https://zenodo.org/records/12784287