###
# i translate every file in csv format
#
###
import csv
import hashlib
import os
import shutil
import time
from os import walk
from config import OP_SYSTEM

HEADERS = ["reboot",
    "malloc_16B_M__1", "malloc_512B_M__1", "malloc_4KB_M__1",
    "malloc_256KB_M__1", "malloc_4MB_M__1", "malloc_128MB_M__1",
    "mmap_huge_M__1","mmap_single_M__1",
    "malloc_16B_M__2", "malloc_512B_M__2", "malloc_4KB_M__2",
    "malloc_256KB_M__2", "malloc_4MB_M__2", "malloc_128MB_M__2",
    "mmap_huge_M__2","mmap_single_M__2",
"malloc_16B_ThA__1", "malloc_512B_ThA__1", "malloc_4KB_ThA__1",
    "malloc_256KB_ThA__1", "malloc_4MB_ThA__1", "malloc_128MB_ThA__1",
    "mmap_huge_ThA__1","mmap_single_ThA__1",
    "malloc_16B_ThA__2", "malloc_512B_ThA__2", "malloc_4KB_ThA__2",
    "malloc_256KB_ThA__2", "malloc_4MB_ThA__2", "malloc_128MB_ThA__2",
    "mmap_huge_ThA__2","mmap_single_ThA__2",
"malloc_16B_ThB__1", "malloc_512B_ThB__1", "malloc_4KB_ThB__1",
    "malloc_256KB_ThB__1", "malloc_4MB_ThB__1", "malloc_128MB_ThB__1",
    "mmap_huge_ThB__1","mmap_single_ThB__1",
    "malloc_16B_ThB__2", "malloc_512B_ThB__2", "malloc_4KB_ThB__2",
    "malloc_256KB_ThB__2", "malloc_4MB_ThB__2", "malloc_128MB_ThB__2",
    "mmap_huge_ThB__2","mmap_single_ThB__2",
    "stack_var_ThA", "tls_var_ThA", "stack_var_ThB", "tls_var_ThB",
    "argv", "env", "stack_var_M", "global_var", "shared_M__1",
    "shared_M__2", "tls_var_M", "lib__1", "lib__2", "text"
]

def ftime(ti):
    return time.strftime('%H:%M:%S', time.gmtime(ti))


def parseFile(source, dst, start, reboot):
    row = {}  # compose the single row of csv

    csvwriter = csv.DictWriter(dst, HEADERS)
    i = 0
    # scorro tutto il file
    while True:
        line = source.readline()
        # se raggiungo la fine esco
        if not line:
            print("EoF reached\n")
            break
        line = " ".join(line.split())  # bonifico la stringa da spazi inutili
        tmp, flag = parse(line)

        if not flag:  # reached the end of sample
            if row:
                row.update({"reboot": reboot})
                csvwriter.writerow(row)
                i += 1
            row = {}
            if not i % 1000:
                print(i, "Elapsed:", ftime(time.time() - start))

        row.update(tmp)  # add the new key to db, this way the order isn't important


def parse(line):
    seg = line.split(" ")  # split the various column

    if len(seg) == 1:  # if found the end of sample, reset counter
        return {}, 0

    # thread segments 1
    obj = seg[0]
    if obj in ["PID", ]:
        return {}, 1
    return {obj: address_split(seg[1])}, 1


def address_split(add):
    return add.split("-")[0].replace("0x", "")


'''
____________---____---___---___---___---___--___---___---__----___--__
'''


def check_hash(data_path):
    archives_path = os.path.join(data_path, "archives")
    n_archives = len(next(walk(archives_path), (None, None, []))[2])  # reading the directory
    if n_archives % 2:
        print(n_archives, "Some files are missing! Check that every archive have its meta.txt file. Exit")
        exit(1)
    n_archives = n_archives // 2

    print("Found archives:", n_archives)

    # check every archive for errors in trasmission
    for i in range(0, n_archives):
        print("Checking raw:", i)
        try:
            raw_fpath = os.path.join(archives_path, f"raw_{i}.zip")
            f = open(raw_fpath, "rb")
        except:
            n_archives+=1
            i+=1
            continue
        hs = hashlib.md5(f.read()).hexdigest()
        f.close()

        meta_fpath = os.path.join(archives_path, f"meta_{i}.txt")
        f = open(meta_fpath, "r")
        my_hash = f.readline().replace("\n", "")
        f.close()

        print("Checking hash: ", hs, my_hash)
        if hs != my_hash:
            my_hash = input("Checksum failed, manually insert hash?")

        if hs != my_hash:
            print(f"Checksum for raw_{i}.zip Failed, Exit")
            exit(1)

    print("ALL Checksum OK")
    return n_archives


def main(data_path):
    archives_path = os.path.join(data_path, "archives")
    if os.path.isdir(archives_path):
        n_archives = check_hash(data_path)

    print("Unpacking Data\n")

    raw_csv_fpath = os.path.join(data_path, 'raw_data.csv')
    dest = open(raw_csv_fpath, "w+")  # destination
    csvwriter = csv.DictWriter(dest, HEADERS)
    csvwriter.writeheader()
    start = time.time()

    '''
    for old windows samples and other data are placed inside another folder called data so replace the path in walk and
    inside the nested for qith /data/data
    '''
    data_dirpath = os.path.join(data_path, "data")
    for i in range(0, n_archives):
        try:
            raw_zip_fpath = os.path.join(archives_path, f"raw_{i}.zip")
            shutil.unpack_archive(raw_zip_fpath, data_dirpath, "zip")
            filenames = next(walk(data_dirpath), (None, None, []))[2]  # [] if no file
        except:
            n_archives+=1
            i+=1
            continue
        print(i, "Parsing Data\n")
        round_time = time.time()

        for name in filenames:
            filepath = os.path.join(data_dirpath, name)
            f = open(filepath, "r")  # source
            parseFile(f, dest, start, i)
            print(i, name, "parsed in:", ftime(time.time() - round_time))
            f.close()
            round_time = time.time()

    dest.close()
    print("Total:", ftime(time.time() - start))


if __name__ == "__main__":
    if OP_SYSTEM == "Linux 5":
        data_path = os.path.join("..", "..", "samples", "data_linux", "5.17.15")
    elif OP_SYSTEM == "Linux 6":
        data_path = os.path.join("..", "..", "samples", "data_linux", "6.4.9")
    elif OP_SYSTEM == "Native":
        data_path = os.path.join("..", "..", "samples", "data_macos", "native", "singleboot")
        reboot_data_path = os.path.join("..", "..", "samples", "data_macos", "native", "reboot")
        main(reboot_data_path)
    elif OP_SYSTEM == "Rosetta":
        data_path = os.path.join("..", "..", "samples", "data_macos", "rosetta", "singleboot")
        reboot_data_path = os.path.join("..", "..", "samples", "data_macos", "rosetta", "reboot")
        main(reboot_data_path)
    elif OP_SYSTEM == "Windows":
        data_path = os.path.join("..", "..", "samples", "data_windows", "singleboot")
        reboot_data_path = os.path.join("..", "..", "samples", "data_windows", "reboot")
        main(reboot_data_path)
    main(data_path)
    
