#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h> // unix standard library
#include <pthread.h> // threading
#include <math.h>
#include <sys/types.h> // system types for linux
#include <sys/mman.h> // used for mmap
#include <linux/mman.h> // used for HUGE TAB size
#include <sys/shm.h>
#include <errno.h>

#define ONE_PAGE (4*1024) //1 page on x86 4 level paging
#define HUGE_PAGE (2*1024*1024)

#define BASE_ALLOC 16 // chunk size

//this one just used to build the command for the two threads, saving us 2 system calls
#define SUPPORT_STRING_SIZE (4000)
char comm[8][SUPPORT_STRING_SIZE];

/* ___________________________________GLOBAL VARIABLES ______________________________*/
//a standard global var. everybody, also threads, see the SAME ADDRESS so we can sample it just
//once and thats it
int global_var;

// a tls variable, its being copied in a private space for threads to use without locking
// and then sync only the writes
__thread int tls_var;


/* ___________________________________ALLOC FUNCTION______________________________*/

void * allocation_func(int i, int multiplicity, char * tag){

    char * malloc_16B = (char *) malloc(BASE_ALLOC);
    char * malloc_512B = (char *) malloc(BASE_ALLOC<<5);
    char * malloc_4KB = (char *) malloc(BASE_ALLOC<<8);
    char * malloc_256KB = (char *) malloc(BASE_ALLOC<<14);
    char * malloc_4MB = (char *) malloc(BASE_ALLOC<<18);
    char * malloc_128MB = (char *) malloc(BASE_ALLOC<<23);


    void * mmap_single = mmap(NULL, ONE_PAGE,  PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, 0);
    void * mmap_huge = mmap(NULL, ONE_PAGE,  PROT_READ|PROT_WRITE|PROT_EXEC, MAP_HUGETLB|MAP_HUGE_2MB|MAP_ANONYMOUS|MAP_PRIVATE, -1, 0);

    if(mmap_huge == MAP_FAILED || mmap_single == MAP_FAILED){
        printf("Something went wrong with mmap()! %s\n", strerror(errno));
        return 1;
    }

    int index = (2*i)+multiplicity-1;
    sprintf(comm[index],
    "malloc_16B_%s__%d %p-0\n"
    "malloc_512B_%s__%d %p-0\n"
    "malloc_4KB_%s__%d %p-0\n"
    "malloc_256KB_%s__%d %p-0\n"
    "malloc_4MB_%s__%d %p-0\n"
    "malloc_128MB_%s__%d %p-0\n"
    "mmap_single_%s__%d %p-0\n"
    "mmap_huge_%s__%d %p-0",
    tag,multiplicity, malloc_16B,
    tag,multiplicity, malloc_512B,
    tag,multiplicity, malloc_4KB,
    tag,multiplicity, malloc_256KB,
    tag,multiplicity, malloc_4MB,
    tag,multiplicity, malloc_128MB,
    tag,multiplicity, mmap_single,
    tag,multiplicity, mmap_huge
    ); // building the command



}

/* ___________________________________THREADS______________________________*/
// We create two threads, A and B, to gather data about the relative positions of objects

void * thread_func(void * arg) {
    // Retrieve a stack address. stack objects are consecutive so we need only one object
    void * stack_var_th;

    char * a = "ThA";
    char * b = "ThB";
    char * tag;

    // Read the naming reference from args. Used to tag the data to their thread (A or B)
    int index = * (int*) arg;
    if( index == 1)
        tag = a;
    else
        tag= b;

    // Make multiple allocation like main
    allocation_func(index,1,tag);
    allocation_func(index,2,tag);

    sprintf(comm[index+5],
    "stack_var_%s %p-0\n"
    "tls_var_%s %p-0",
    tag, &stack_var_th,
    tag, &tls_var
    ); // building the command
}

/* ___________________________________MAIN_________________________________*/

int main(int argc, char *argv[]) {
    char * tag = "M";
    void * stack_var_main;

//1) get the overrall virtual memory mapping and save that

    int pid = getpid(); // getting the pid of the current process

// 2) shared memory segment
     int shm_id_1 = shmget(IPC_PRIVATE, ONE_PAGE, IPC_CREAT |  0666 );
     int shm_id_2 = shmget(IPC_PRIVATE, ONE_PAGE, IPC_CREAT |  0666 );

     if (shm_id_1 < 0 || shm_id_2 <0) {
     printf("shmget error\n %s", strerror(errno));
     exit(1);
    }

     void * shared_mem_1 = shmat(shm_id_1,0,0);
     void * shared_mem_2 = shmat(shm_id_2,0,0);

     if ( shared_mem_1 < 0 || shared_mem_2 <0) {
     printf("shmmat error\n %s", strerror(errno));
     exit(1);
    }


// 3) MALLOC & MMAP ZONE

    allocation_func(0,1,tag);
    allocation_func(0,2,tag);



// 4) get env and argv addresses

    char * env = getenv("PATH");

// 5) thread

    pthread_t thread_1,thread_2;
    void * thread_1_status, * thread_2_status ;

    int first = 1, second = 2;
    pthread_create(&thread_1, NULL, thread_func, &first);
    pthread_create(&thread_2, NULL, thread_func, &second);

    pthread_join(thread_1, &thread_1_status);
    pthread_join(thread_2, &thread_2_status);


//Build the command
    char a[2] ="a";
    char b[2] ="b";
    strncpy(a,b, 2);

    printf("#%s\n"
        "%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n" // Allocation information
        "argv %p-0\n"
        "env %p-0\n"
        "stack_var_M %p-0\n"
        "global_var %p-0\n"
        "shared_M__1 %p-0\n"
        "shared_M__2 %p-0\n"
        "tls_var_M %p-0\n"
        "lib__1 %p-0\n"
        "lib__2 %p-0\n"
        "text %p-0\n"
        "PID %d\n",
        argv[1], // file counter used to multithread sampling
        comm[0],comm[1], // allocation from MAIN
        comm[2],comm[3],comm[6], // allocation from ThA
        comm[4],comm[5],comm[7], // allocation from ThB
        argv, //ArgV Address
        env, //env variable Address
        &stack_var_main,
        &global_var,
        shared_mem_1,
        shared_mem_2,
        &tls_var,
        &malloc,
        &log2f,
        &main,
        pid
        );

        if(argv[1]==NULL){
            char test;
            scanf("%c",&test);
        }

        int err=0;
        err += shmdt(shared_mem_1);
        err += shmdt(shared_mem_2);
        err += shmctl(shm_id_1, IPC_RMID, NULL);
        err += shmctl(shm_id_2, IPC_RMID, NULL);

        if ( err < 0 ) {
            printf("shmdt error\n %s", strerror(errno));
            exit(1);
        }


    return 0;
}
