//#include <stdlib.h>
//#include <string.h>
//#include <sys/types.h> // system types for linux
//#include <stddef.h>
//#include <stdint.h>
//#include <errno.h>
//windows specific
#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <conio.h>
#include <process.h>
#include <tchar.h>

#define PAGE_SIZE (1024*4) // Windows 4KB small page size
#define HUGE_PAGE_SIZE (2*1024*1024) // 2097152 = 2MB obtained by GetLargePageMinimum()

// threshold are seemed to be variable, this value look reliable
#define BASE_ALLOC 16 // chunk size

#define MMAP_ALLOC_SIZE (PAGE_SIZE) //1 page on on Windows
#define SHARED_SIZE (PAGE_SIZE) //4KB so 1 page on Windows

//this one just used to build the command for the two threads, saving us 2 system calls
#define SUPPORT_STRING_SIZE (4000)
char comm[8][SUPPORT_STRING_SIZE];

typedef HRESULT (CALLBACK* LPFNDLLFUNC1)(DWORD,UINT*);
typedef HRESULT (CALLBACK* GLE)();
typedef HRESULT (CALLBACK* LUPV)(LPCWSTR,LPCWSTR,PLUID);
typedef HRESULT (CALLBACK* OPT)(HANDLE,DWORD,PHANDLE);
typedef HRESULT (CALLBACK* ATP)(HANDLE,BOOL,PTOKEN_PRIVILEGES,DWORD,PTOKEN_PRIVILEGES,PDWORD);

/* ___________________________________Inc Priviledge FUNCTION______________________________*/

void Privilege(const wchar_t* pszPrivilege, BOOL bEnable)
{
    HANDLE           hToken;
    TOKEN_PRIVILEGES tp;
    BOOL             status;
    DWORD            error;

    HINSTANCE dll_handler;               // Handle to DLL
    HINSTANCE dll_handler2;               // Handle to DLL

    OPT OpenProcessToken;    // Function pointer
    ATP AdjustTokenPrivileges;    // Function pointer
    LUPV LookupPrivilegeValue;    // Function pointer
    GLE GetLastError;    // Function pointer

    dll_handler = LoadLibrary("Advapi32");
    dll_handler2 = LoadLibrary("Kernel32");
    OpenProcessToken = (OPT)GetProcAddress(dll_handler, "OpenProcessToken");
    AdjustTokenPrivileges = (ATP)GetProcAddress(dll_handler, "AdjustTokenPrivileges");
    LookupPrivilegeValue = (LUPV)GetProcAddress(dll_handler, "LookupPrivilegeValueA");
    GetLastError = (GLE)GetProcAddress(dll_handler2, "GetLastError");

    // open process token
    if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken))
        printf("error: Something went wrong priviledge! %s\n", strerror(errno));

    // get the luid
    if (!LookupPrivilegeValue(NULL, pszPrivilege, &tp.Privileges[0].Luid))
        printf("error: Something went wrong priviledge! %s\n", strerror(errno));
    tp.PrivilegeCount = 1;

    // enable or disable privilege
    if (bEnable)
        tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
    else
        tp.Privileges[0].Attributes = 0;

    // enable or disable privilege
    status = AdjustTokenPrivileges(hToken, FALSE, &tp, 0, (PTOKEN_PRIVILEGES)NULL, 0);

    // It is possible for AdjustTokenPrivileges to return TRUE and still not succeed.
    // So always check for the last error value.
    error = GetLastError();
    if (!status || (error != ERROR_SUCCESS))
        printf("error: Something went wrong priviledge! %d\n", error);

    // close the handle
    if (!CloseHandle(hToken))
        printf("error: Something went wrong priviledge! %s\n", strerror(errno));
}


/* ___________________________________GLOBAL VARIABLES ______________________________*/
//a standard global var. everybody, also threads, see the SAME ADDRESS so we can sample it just
//once and thats it
int global_var;

// a tls variable, its being copied in a private space for threads to use without locking
// and then sync only the writes
__declspec(thread) int tls_var;

/* ___________________________________ALLOC FUNCTION______________________________*/

void * allocation_func(int i, int multiplicity, char * tag){

    char * malloc_16B = (char *) malloc(BASE_ALLOC);
    char * malloc_512B = (char *) malloc(BASE_ALLOC<<5);
    char * malloc_4KB = (char *) malloc(BASE_ALLOC<<8);
    char * malloc_256KB = (char *) malloc(BASE_ALLOC<<14);
    char * malloc_4MB = (char *) malloc(BASE_ALLOC<<18);
    char * malloc_128MB = (char *) malloc(BASE_ALLOC<<23);

    void * virtalloc = VirtualAlloc(NULL, MMAP_ALLOC_SIZE, MEM_COMMIT|MEM_RESERVE, PAGE_READWRITE );

    Privilege(TEXT("SeLockMemoryPrivilege"), TRUE);
    void * virtalloc_huge = VirtualAlloc(NULL, HUGE_PAGE_SIZE, MEM_COMMIT|MEM_RESERVE|MEM_LARGE_PAGES, PAGE_READWRITE );

    if(!virtalloc){
        printf("error: Something went wrong with VirtualAlloc()! %s\n", strerror(errno));
        return 1;
    }
    if(!virtalloc_huge){
        printf("error: Something went wrong with VirtualAlloc() of huge page! %s\n", strerror(errno));
        return 1;
    }

     int index = (2*i)+multiplicity-1;
    sprintf(comm[index],
    "malloc_16B_%s__%d %p-0\n"
    "malloc_512B_%s__%d %p-0\n"
    "malloc_4KB_%s__%d %p-0\n"
    "malloc_256KB_%s__%d %p-0\n"
    "malloc_4MB_%s__%d %p-0\n"
    "malloc_128MB_%s__%d %p-0\n"
    "mmap_single_%s__%d %p-0\n"
    "mmap_huge_%s__%d %p-0",
    tag,multiplicity, malloc_16B,
    tag,multiplicity, malloc_512B,
    tag,multiplicity, malloc_4KB,
    tag,multiplicity, malloc_256KB,
    tag,multiplicity, malloc_4MB,
    tag,multiplicity, malloc_128MB,
    tag,multiplicity, virtalloc,
    tag,multiplicity, virtalloc_huge
    ); // building the command
}

/* ___________________________________THREADS______________________________*/
// We create two threads, A and B, to gather data about the relative positions of objects

void * thread_func(void * arg) {
    // Retrieve a stack address. stack objects are consecutive so we need only one object
    void * stack_var_th;

    char * a = "ThA";
    char * b = "ThB";
    char * tag;

    // Read the naming reference from args. Used to tag the data to their thread (A or B)
    int index = * (int*) arg;
    if( index == 1)
        tag = a;
    else
        tag= b;

    // Make multiple allocation like main
    allocation_func(index,1,tag);
    allocation_func(index,2,tag);

    sprintf(comm[index+5],
    "stack_var_%s %p-0\n"
    "tls_var_%s %p-0",
    tag, &stack_var_th,
    tag, &tls_var
    ); // building the command
}

/* ___________________________________MAIN_________________________________*/

int main(int argc, char *argv[]) {
    char * tag = "M";
    void * stack_var_main;

//1) get the overrall virtual memory mapping and save that

    int pid = getpid(); // getting the pid of the current process

// 2) shared memory segment
    HANDLE shm_id_1,shm_id_2;
    LPCTSTR shared_mem_1, shared_mem_2;
    TCHAR szName1[]=TEXT("MyFileMappingObject1");
    TCHAR szName2[]=TEXT("MyFileMappingObject2");

    shm_id_1 = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE | SEC_COMMIT, 0, SHARED_SIZE, szName1);                 // name of mapping object
    shm_id_2 = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE | SEC_COMMIT, 0, SHARED_SIZE, szName2);                 // name of mapping object

    if (shm_id_1 == NULL || shm_id_2 == NULL){
        printf("error: CreateFileMapping Error %d", GetLastError());
        exit(-2);
    }

   shared_mem_1 = (LPTSTR) MapViewOfFile(shm_id_1, FILE_MAP_ALL_ACCESS, 0, 0, SHARED_SIZE);
   shared_mem_2 = (LPTSTR) MapViewOfFile(shm_id_2, FILE_MAP_ALL_ACCESS, 0, 0, SHARED_SIZE);

   if (shared_mem_1 == NULL || shared_mem_2 == NULL){
      printf("error: Could not map view of file (%d).\n", GetLastError());
      CloseHandle(shm_id_1);
      return 1;
   }


// 3) MALLOC & MMAP ZONE

    allocation_func(0,1,tag);
    allocation_func(0,2,tag);



// 4) get env and argv addresses

    char * env = getenv("PATH");

// 5) thread
    uintptr_t thread_1,thread_2;

    int first = 1, second = 2;

    thread_1 = _beginthread(thread_func,0,&first);
    thread_2 = _beginthread(thread_func,0,&second);

    WaitForSingleObject((HANDLE)thread_1, INFINITE);
    WaitForSingleObject((HANDLE)thread_2, INFINITE);

// 6) load a DLL explicitly
    HINSTANCE dll_handler;               // Handle to DLL
    LPFNDLLFUNC1 MD5Init;    // Function pointer

    dll_handler = LoadLibrary("ntdll");
    MD5Init = (LPFNDLLFUNC1)GetProcAddress(dll_handler, "MD5Init");

//Build the command

    printf("#%s\n"
        "%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n" // Allocation information
        "argv %p-0\n"
        "env %p-0\n"
        "stack_var_M %p-0\n"
        "global_var %p-0\n"
        "shared_M__1 %p-0\n"
        "shared_M__2 %p-0\n"
        "tls_var_M %p-0\n"
        "lib__1 %p-0\n"
        "lib__2 %p-0\n"
        "text %p-0\n"
        "PID %d\n",
        argv[1], // file counter used to multithread sampling
        comm[0],comm[1], // allocation from MAIN
        comm[2],comm[3],comm[6], // allocation from ThA
        comm[4],comm[5],comm[7], // allocation from ThB
        argv, //ArgV Address
        env, //env variable Address
        &stack_var_main,
        &global_var,
        shared_mem_1,
        shared_mem_2,
        &tls_var,
        MD5Init,
        &CreateFileMapping,
        &main,
        pid
        );

// debug

if(argv[1]==NULL){
    char test;
    scanf("char %c",&test);
}

//closing stuff
    UnmapViewOfFile(shared_mem_1);
    UnmapViewOfFile(shared_mem_2);
    CloseHandle(shm_id_1);
    CloseHandle(shm_id_2);

    return 0;
}
