from .universal_sampler import *


class LinuxSampler(Sampler):
    def __init__(self):
        super().__init__(platform.system())

    def __compile_executable__(self):
        print('compiling c code')
        if subprocess.run(["gcc", "-pthread", "-w", self.exec_path + ".c", "-o", self.exec_path, "-lm"]).returncode:
            print("something went wrong with the compilation process")
            exit()

    def __activate_hugepages__(self):
        if True and subprocess.run(["sudo", "-S", "sysctl", "-w", "vm.nr_hugepages=512"]).returncode:
            print("something went wrong with the allocation of hugepages")
            exit()

    def __check_aslr__(self):
        # check if ASLR active:
        '''
        0 – No randomization. Everything is static.
        1 – Conservative randomization. Shared libraries, stack, mmap(),
            VDSO and heap are randomized.
        2 – Full randomization. In addition to elements listed in the previous point,
            memory managed through brk() is also randomized.
        '''
        aslr = subprocess \
            .Popen(["cat", "/proc/sys/kernel/randomize_va_space"], stdout=subprocess.PIPE) \
            .communicate()[0].decode()

        print("ASLR level:", aslr)
        if "2" not in aslr:
            print("increasing ASLR level")
            if subprocess.run(["sudo", "-S", "sysctl", "-w", "kernel.randomize_va_space=2"]).returncode:
                print("something went wrong when increasing ALSR level")
                exit()

    def __check_pie__(self):
            # check that the executable is indeed PIE

            pie = subprocess\
                .Popen(["file", "./" + self.exec_path], stdout=subprocess.PIPE)\
                .communicate()[0].decode()
            pie = pie.split(",")[0]
            print("executable type:", pie)
            if "pie" not in pie:
                print("The executable is not PIE! check compiler flags")
                exit()
            print("OK its PIE")

            print("OK its PIE and compiled for the right architecture")

    def initialize_all(self):
        self.__compile_executable__()
        self.__activate_hugepages__()
        self.__check_aslr__()
        self.__check_pie__()
