import subprocess
from .universal_sampler import *


class MacosSampler(Sampler):
    def __init__(self, rosetta=False):
        super().__init__(platform.system())
        self.rosetta = rosetta
        self.src = self.exec_path



    def __compile_executable__(self):
        # compiling the executable

        if self.rosetta:
            print('compiling c code for INTEL')
            cmd = ["gcc", "-pthread", "-w", self.src + ".c", "-o", self.exec_path]
        else:
            print('compile the executable yourself in the terminal')
            print(f"gcc -pthread -w {self.exec_path}.c -o {self.exec_path}")

        if self.rosetta and subprocess.run(cmd).returncode:
            print("something went wrong with the compilation process")
            exit()

    def __check_pie__(self):
        # check that the executable is indeed PIE and compiled for right architecture
        details = subprocess\
            .Popen(["otool", "-hv", "./" + self.exec_path], stdout=subprocess.PIPE)\
            .communicate()[0].decode()
        print("executable type:", details)
        if self.rosetta and "X86_64" not in details:
            print("The executable is not compiled for INTEL!")
            exit()

        if not self.rosetta and "ARM64" not in details:
            print("The executable is not compiled for APPLE SILICON!")
            exit()

        if "PIE" not in details:
            print("The executable is not PIE! check compiler flags")
            exit()

        print("OK its PIE and compiled for the right architecture")

    def set_rosetta(self):
        self.rosetta = True

        print("ROsETTA")
        self.working_dir = self.working_dir + "/ROSETTA"  # eg ./data_linux
        self.data_dir = self.working_dir + DATA_PATH  # eg ./data_linux/data/
        self.archive_dir = self.working_dir + ARCHIVE_PATH  # eg ./data_linux/archives/
        self.exec_path = EXECUTABLES_PATH + "/" + self.exec_name + "_ROSETTA"

        # check directory
        if not os.path.isdir("./" + self.working_dir):
            os.mkdir("./" + self.working_dir)
        if not os.path.isdir("./" + self.data_dir):
            os.mkdir("./" + self.data_dir)
        if not os.path.isdir("./" + self.archive_dir):
            os.mkdir("./" + self.archive_dir)

    def ask_initial_param(self):
        val = input(f"Want to use ROSETTA?t/f: ({self.rosetta})") or self.rosetta

        if "t" in str(val):
            self.set_rosetta()
        print("rosetta:", self.rosetta)
        super().ask_initial_param()

    def quick_init(self, all_sample, thread_number, reboot_counter, rosetta):
        super().quick_init(all_sample, thread_number, reboot_counter)
        if int(rosetta):
            self.set_rosetta()


    def initialize_all(self):

        self.__compile_executable__()
        self.__check_pie__()

