from .universal_sampler import *


class WindowsSampler(Sampler):
    def __init__(self):
        super().__init__(platform.system())
        self.recompile_flag = True

    def __compile_executable__(self):

        temp = EXECUTABLES_PATH.replace("/", "\\")
        cmd = ["cl", "/Fe"+temp+self.exec_name+".exe", "/Fo"+temp+self.exec_name+".obj", self.exec_path + ".c", "-w"]

        if self.recompile_flag and subprocess.run(cmd).returncode:
            print("something went wrong with the compilation process")
            exit()
        print('______________________ COMPILED OK ____________________________________\n\n')

    def __check_pie__(self):
        # check that the executable is indeed PIE and compiled for right architecture

        file_headers = \
            subprocess.Popen(["dumpbin", self.exec_path + ".exe", "/headers"], stdout=subprocess.PIPE).communicate()[
                0].decode()

        if "machine (x64)" not in file_headers:
            print("The executable is not 64 bit: compile with 'x64 Native Tools Command Prompt' terminal")
            exit()

        if "Dynamic base" not in file_headers:
            print("ASLR is not active, check 'Exploit Protection' Settings")
            exit()

        if "High Entropy Virtual Addresses" not in file_headers:
            print("ASLR  'High Entropy' is not active, check 'Exploit Protection' Settings")
            exit()

        print("OK its ASLR-HighEntropy and compiled for the right architecture")

    def quick_init(self, all_sample, thread_number, reboot_counter):
        super().quick_init(all_sample, thread_number, reboot_counter)
        self.recompile_flag = False

    def initialize_all(self):
        self.__compile_executable__()
        self.__check_pie__()
